var AnalyticsReporter = require('../../lib/transactions/analytics-reporter').AnalyticsReporter
  ,   ConfigManager = require('../../lib/transactions/config-manager').ConfigManager
  ,   mocks = require('../helpers/mocks')
  ,   http = require('http');

describe('AnalyticsReporter', function() {
  var agentMock, configManager, reporter;

  beforeEach(function() {
    agentMock = new mocks.AgentMock();
    agentMock.opts = {
      applicationName: 'test-app',
      tierName: 'test-tier',
      nodeName: 'test-node',
      analytics: {
        host: 'test-host',
        port: 9876
      }
    };
    agentMock.appdNative = {
      initAnalytics: jasmine.createSpy('init analytics') ,
      enableAnalytics: jasmine.createSpy('enable analytics'),
      recordTransaction: jasmine.createSpy('record transaction'),
      reportTransactions: jasmine.createSpy('report analytics')
    };

    configManager = new ConfigManager(agentMock);
    agentMock.configManager = configManager;
    configManager.config = {};

    reporter = new AnalyticsReporter(agentMock);
  });

  describe('defaults', function() {
    it('defaults analytics agent host and port', function() {
      expect(reporter.analyticsHost).toEqual('localhost');
      expect(reporter.analyticsPort).toEqual(9090);
    });
  });

  describe('#init()', function() {
    it('uses default host and port if not configured', function() {
      agentMock.opts = {};
      reporter.init();
      expect(reporter.analyticsHost).toEqual('localhost');
      expect(reporter.analyticsPort).toEqual(9090);
    });
    it('reads host and port from config', function() {
      agentMock.opts = {
        analytics: {
          host: 'example.com',
          port: 1234
        }
      };
      reporter.init();
      expect(reporter.analyticsHost).toEqual('example.com');
      expect(reporter.analyticsPort).toEqual(1234);
    });
    it('sets up app config listener', function() {
      var configUpdated = spyOn(reporter, 'configUpdated');
      var initNative = spyOn(reporter, 'initNative');

      reporter.init();

      expect(configUpdated).toHaveBeenCalled();
      expect(initNative).not.toHaveBeenCalled();
      agentMock.emit('configUpdated');
      expect(configUpdated).toHaveBeenCalled();
      expect(initNative).not.toHaveBeenCalled();
    });
  });

  describe('#initNative', function() {
    var configUpdated;

    function setAgentIdentity() {
      agentMock.configManager.config = {
        agentIdentity: {
          appID: 'app-id',
          tierID: 'tier-id',
          nodeID: 'node-id'
        }
      };
    }

    beforeEach(function() {
      configUpdated = spyOn(reporter, 'configUpdated');
    });

    it('does nothing if agent identity is not yet known', function() {
      reporter.initNative();
      expect(agentMock.appdNative.initAnalytics).not.toHaveBeenCalled();
    });
    it('calls native delegate once agent identity is known', function() {
      setAgentIdentity();

      reporter.init();
      reporter.initNative();
      agentMock.timers.clearInterval(reporter._reportInterval);

      expect(reporter.initialized).toBe(true);
      expect(agentMock.appdNative.initAnalytics).toHaveBeenCalled();
      expect(agentMock.appdNative.initAnalytics.calls.argsFor(0)[0]).toEqual({
        applicationName: 'test-app',
        tierName: 'test-tier',
        nodeName: 'test-node',
        applicationId: 'app-id',
        tierId: 'tier-id',
        nodeId: 'node-id'
      });
    });
    it('initializes analytics configuration', function() {
      setAgentIdentity();

      reporter.init();
      reporter.initNative();
      agentMock.timers.clearInterval(reporter._reportInterval);

      expect(configUpdated).toHaveBeenCalled();
    });
    it('installs anaylitics reporting handler', function() {
      var analyticsData = {};
      agentMock.appdNative.initAnalytics = function(config, cb) {
        cb(analyticsData);
      };
      var request = {
        on: jasmine.createSpy('request.on'),
        write: jasmine.createSpy('request.write'),
        end: jasmine.createSpy('request.write')
      };
      http.request = jasmine.createSpy('http.request').and.callFake(function() {
        return request;
      });

      setAgentIdentity();

      reporter.init();
      reporter.initNative();
      agentMock.timers.clearInterval(reporter._reportInterval);

      expect(http.request).toHaveBeenCalledWith({
        appdIgnore: true,
        host: 'test-host',
        port: 9876,
        path: '/v1/sinks/bt',
        method: 'POST',
        headers: {
          'Content-Type': 'application/json'
        }
      });
      expect(request.on).toHaveBeenCalled();
      expect(request.on.calls.count()).toBe(2);
      expect(request.on.calls.argsFor(0)[0]).toBe('response');
      expect(request.on.calls.argsFor(1)[0]).toBe('error');
      expect(request.write).toHaveBeenCalledWith(analyticsData);
      expect(request.end).toHaveBeenCalled();
    });
  });

  describe('#configUpdate()', function() {
    beforeEach(function() {
      configManager.config = {
        analyticsConfig: {
          isEnabled: true,
          analyticsBTIDs: [1,2,3]
        }
      };
    });

    afterEach(function() {
      configManager.config.analyticsConfig.isEnabled = false;
      reporter.enabled = true;
      reporter.configUpdated();
    });

    it('initializes analytics if needed', function() {
      var initNative = spyOn(reporter, 'initNative');

      reporter.initialized = false;
      reporter.configUpdated();
      expect(initNative).toHaveBeenCalled();
    });
    it('doesn\'t re-initialize analytics', function() {
      var initNative = spyOn(reporter, 'initNative');

      reporter.initialized = true;
      reporter.configUpdated();
      expect(initNative).not.toHaveBeenCalled();
    });
    it('enables analytics when turned on', function() {
      reporter.enabled = false;
      configManager.config.analyticsConfig.isEnabled = true;
      reporter.configUpdated();
      expect(agentMock.appdNative.enableAnalytics).toHaveBeenCalledWith(true);
    });
    it('doesn\'t re-enable analytics when still turned on', function() {
      reporter.enabled = true;
      configManager.config.analyticsConfig.isEnabled = true;
      reporter.configUpdated();
      expect(agentMock.appdNative.enableAnalytics).not.toHaveBeenCalled();
    });
    it('records analytics enabled BT IDs', function() {
      reporter.configUpdated();
      expect(reporter.enabledBTIDs).toEqual([1,2,3]);
    });
    it('sets up monitors on enable analytics', function() {
      reporter.configUpdated();
      expect(reporter._transactionHandler).not.toBeNull();
      expect(reporter._reportInterval).not.toBeNull();
    });
    it('removes monitors on disable analytics', function() {
      reporter.configUpdated();
      configManager.config.analyticsConfig.isEnabled = false;
      reporter.configUpdated();
      expect(reporter._transactionHandler).toBeNull();
      expect(reporter._reportInterval).toBeNull();
    });
  });

  describe('#getTimestamp()', function() {
    it('returns correctly formatted date', function() {
      expect(/\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{3}Z/.exec(reporter.getTimestamp()));
    });
  });

  describe('#addTransaction()', function() {
    beforeEach(function() {
      configManager.config = {
        analyticsConfig: {
          isEnabled: true,
          analyticsBTIDs: ['1']
        }
      };
      reporter.init();
    });

    afterEach(function() {
      configManager.config.analyticsConfig.isEnabled = false;
      reporter.configUpdated();
    });

    it('does nothing if analytics is disabled', function() {
      // update enabled via configUpdated() so monitors get removed
      configManager.config.analyticsConfig.isEnabled = false;
      reporter.configUpdated();
      reporter.addTransaction({});
      expect(agentMock.appdNative.recordTransaction).not.toHaveBeenCalled();
    });
    it('does nothing if transaction is ignored', function() {
      reporter.addTransaction({ ignore: true });
      expect(agentMock.appdNative.recordTransaction).not.toHaveBeenCalled();
    });
    it('does nothing if transaction is not yet registered', function() {
      reporter.addTransaction({ registrationId: null });
      expect(agentMock.appdNative.recordTransaction).not.toHaveBeenCalled();
    });
    it('does nothing if transaction is not analytics enabled', function() {
      reporter.addTransaction({ registrationId: 999 });
      expect(agentMock.appdNative.recordTransaction).not.toHaveBeenCalled();
    });
    it('calls native recorder', function() {
      var date = (new Date()).toISOString();
      spyOn(reporter, 'getTimestamp').and.returnValue(date);

      var txn = {
        guid: 1,
        registrationId: 1,
        name: 'test',
        ms: 3,
        eumGuid: 4,
        exitCalls: [{
          exitType: 'EXIT_TEST',
          ms: 9,
        }],
        httpRequestAnalyticsData: {
          cookies: 'cookies',
          headers: 'headers',
          httpParams: 'http params'
        },
        url: '/test',
        api: {
          analyticsData: 'analytics data'
        }
      };

      reporter.addTransaction(txn);

      expect(agentMock.appdNative.recordTransaction).toHaveBeenCalledWith({
        eventTimestamp: date,
        requestGUID: 1,
        transactionId: 1,
        transactionName: 'test',
        transactionTime: 3,
        clientRequestGUID: 4,
        requestExperience: 'NORMAL',
        entryPoint: true,
        exitCalls: [{
          exitCallType: 'EXIT_TEST',
          avgResponseTimeMillis: 9,
          numberOfErrors: 0,
          numberOfCalls: 1
        }],
        httpData: {
          cookies: 'cookies',
          headers: 'headers',
          parameters: 'http params',
          url: '/test'
        },
        userData: 'analytics data'
      });
    });
  });

  describe('#reportAnalytics()', function() {
    it('does nothing in not enabled', function() {
      reporter.enabled = false;
      reporter.reportAnalytics();
      expect(agentMock.appdNative.reportTransactions).not.toHaveBeenCalled();
    });
    it('calls through to native reporter', function() {
      reporter.enabled = true;
      reporter.reportAnalytics();
      expect(agentMock.appdNative.reportTransactions).toHaveBeenCalled();
    });
  });

  describe('#getUserExperience()', function() {
    it('detects ERROR state', function() {
      expect(reporter._getUserExperience({
        hasErrors: true
      })).toBe('ERROR');
    });
    it('detects VERY SLOW state', function() {
      expect(reporter._getUserExperience({
        btInfoResponse: {
          currentVerySlowThreshold: 10
        },
        ms: 20
      })).toBe('VERY_SLOW');
    });
    it('detects SLOW state', function() {
      expect(reporter._getUserExperience({
        btInfoResponse: {
          currentSlowThreshold: 10
        },
        ms: 20
      })).toBe('SLOW');
    });
    it('detects NORMAL state', function() {
      expect(reporter._getUserExperience({})).toBe('NORMAL');
    });
  });
});
